/* Branch Predictor Implementation.
 *                                 Written by Kenji KISE, 2003-04-03
 ********************************************************************/
#include "define.h"

/************************************************************/
#define BUF_SIZE 4096
#define B_TKN 1 /* Branch Taken   */
#define B_UNT 0 /* Branch Untaken */

class BP{ /*** Branch Prediction ***/
  int size; /* Hardware budget in bit */
  char name[BUF_SIZE];
  uint64_t num_of_hit, num_of_mis;
public:
  BP();  /* Constructor */
  ~BP(); /* Destructor  */ 
  void set_config(char*, int);
  void bookkeep(int, int);
};

BP::BP(){num_of_hit = num_of_mis = 0;}

BP::~BP(){
  printf("===\n=== BP_Name: %s, ", name);
  printf("%9.2f KB\n", (float)size/8192.0);
  printf("=== Sum HIT MISS HITrate: ");
  printf("%011lld: %011lld %011lld: ",
	 num_of_hit + num_of_mis, 
	 num_of_hit, num_of_mis);
  printf("%06.2f\n",
	 (double)num_of_hit * 100.0 /
	 (double)(num_of_hit + num_of_mis));
}

void BP::set_config(char *n, int s){
  size = s;
  strncpy(name, n, BUF_SIZE);
}

void BP::bookkeep(int pred, int result){
  if(pred==result) num_of_hit++;
  else             num_of_mis++;
}
/************************************************/

/************************************************/
class TwoSC{ /** Two-bit Saturating Counter **/
  char cnt;
public:
  void init(char);
  void increment();
  void decrement();
  char load();
  int predict();
};

void TwoSC::init(char init){ cnt = init; }
char TwoSC::load(){ return cnt;};
void TwoSC::increment(){ if(cnt<3) cnt++; }
void TwoSC::decrement(){ if(cnt>0) cnt--; }
int TwoSC::predict(){
  if(cnt>1) return B_TKN; else return B_UNT;
}
/************************************************/

/*************************************************/
class bimodal : public BP{
  int n;       /* n-bit                 */
  int entry;   /* PHT entry             */
  TwoSC *pht;  /* Pattern History Table */
public:
  bimodal(int); /* Constructor */
  void init();  /* Initialize  */
  void predict(uint64_t, int*);
  void regist(uint64_t, int, int);
};

bimodal::bimodal(int n_bit){
  n     = n_bit;
  entry = 1 << n_bit;
  pht = new TwoSC[entry];
  char buf[BUF_SIZE];
  sprintf(buf, "bimodal(%2d)", n_bit);
  set_config(buf, entry*2);
  init();
}

void bimodal::init(){
  for(int i=0; i<entry; i++) 
    pht[i].init(2); /* Init by Weakly Taken*/
}

void bimodal::predict(uint64_t pc, int *pred){
  int index = (pc>>2) % entry;
  *pred = pht[index].predict();
}

void bimodal::regist(uint64_t pc, int r,
		     int pred){
  int index = (pc>>2) % entry;
  if(r==B_TKN) pht[index].increment();
  if(r==B_UNT) pht[index].decrement();
  bookkeep(r, pred);
}

/************************************************/
class gshare : public BP{
  int n,m;       /* n-bit & m-bit         */
  int entry;     /* PHT entry             */
  TwoSC *pht;    /* Pattern History Table */
 public:
  gshare(int,int); /* Constructor */
  void init();     /* Initialize  */
  void predict(uint64_t, uint64_t, int*);
  void regist(uint64_t, uint64_t, int, int);
};

gshare::gshare(int n_bit, int m_bit){
  n = n_bit;
  m = m_bit;
  entry = 1 << n_bit;
  if(m>n){printf("*** Error\n"); exit(1);}
  pht = new TwoSC[entry];
  char buf[BUF_SIZE];
  sprintf(buf, "gshare(%2d,%2d)", n_bit, m_bit);
  set_config(buf, entry*2);
  init();
}

void gshare::init(){
  for(int i=0; i<entry; i++) 
    pht[i].init(2); /* Init by Weakly Taken*/
}

void gshare::predict(uint64_t pc, uint64_t bhr, 
		     int *pred){
  int index = ((pc>>2) ^ (bhr<<(n-m))) % entry;
  *pred = pht[index].predict();
}

void gshare::regist(uint64_t pc, uint64_t bhr, 
		    int r, int pred){
  int index = ((pc>>2) ^ (bhr<<(n-m))) % entry;
  if(r==B_TKN) pht[index].increment();
  if(r==B_UNT) pht[index].decrement();
  bookkeep(r, pred);
}

/************************************************/
class bimode : public BP{
  int n, m, s, n_entry, s_entry;
  TwoSC *choice;        /* choice predictor */
  TwoSC *pht_t;         /* Taken   PHT      */
  TwoSC *pht_u;         /* Untaken PHT      */
 public:
  bimode(int, int, int); /* Constructor */
  void init();           /* Initialize  */
  void predict(uint64_t, uint64_t, int*);
  void regist(uint64_t, uint64_t, int, int);
};

bimode::bimode(int n_bit, int m_bit, int s_bit){
  n = n_bit;
  m = m_bit;
  s = s_bit;
  n_entry = 1 << n_bit;
  s_entry = 1 << s_bit;
  choice= new TwoSC[s_entry];
  pht_t = new TwoSC[n_entry];
  pht_u = new TwoSC[n_entry];
  char buf[BUF_SIZE];
  sprintf(buf, "bi-mode(%2d,%2d,%2d)", n,m,s);
  set_config(buf, s_entry*2 + n_entry*4);
  init();
}

void bimode::init(){
  for(int i=0; i<n_entry; i++){
    pht_t[i].init(2); /* Weakly Taken   */
    pht_u[i].init(1); /* Weakly UnTaken */
  }
  for(int i=0; i<s_entry; i++)
    choice[i].init(2);/* Weakly Taken   */
}

void bimode::predict(uint64_t pc, uint64_t bhr, 
		     int *pred){
  int index_c = (pc>>2) % s_entry;
  int index_p =((pc>>2) ^ (bhr<<(n-m))) % n_entry;

  if(choice[index_c].predict())
    *pred = pht_t[index_p].predict();
  else
    *pred = pht_u[index_p].predict();
}

void bimode::regist(uint64_t pc, uint64_t bhr, 
		    int r, int pred){
  int index_c = (pc>>2) % s_entry;
  int index_p =((pc>>2) ^ (bhr<<(n-m))) % n_entry;
  
  if(choice[index_c].predict()){
    if(r==B_TKN) pht_t[index_p].increment();
    if(r==B_UNT) pht_t[index_p].decrement();
  }else{
    if(r==B_TKN) pht_u[index_p].increment();
    if(r==B_UNT) pht_u[index_p].decrement();
  }
  
  /** Partial update of the choice PHT **/
  TwoSC *cnt = &choice[index_c]; 
  if(r==cnt->predict() || r!=pred){
    if(r==B_TKN) cnt->increment();
    if(r==B_UNT) cnt->decrement();
  }
  bookkeep(r, pred);
}
/************************************************/

#include "branch_evaluate.cc"
